import argparse
import pathlib
import sys

import tomlkit
from yamk.make import MakeCommand

sys.tracebacklimit = 0


def get_version():
    pyproject = pathlib.Path(__file__).parents[2].joinpath("pyproject.toml")
    with open(pyproject) as toml_file:
        conf = tomlkit.parse(toml_file.read())
    return conf["tool"]["poetry"]["version"]


def parse_args():
    parser = argparse.ArgumentParser(
        prog="yamk", description="Yet another make command",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {get_version()}",
        help="Print the version and exit",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increae the level of verbosity",
    )
    parser.add_argument(
        "-m",
        "--makefile",
        metavar="Makefile",
        default="make.toml",
        help="The path to makefile",
    )
    parser.add_argument(
        "-x",
        "--variable",
        action="append",
        metavar="KEY=value",
        dest="variables",
        default=[],
        help="A list of variables to override the ones set in the makefile. "
        "They should be in the form <variable>=<value>",
    )
    parser.add_argument(
        "target", nargs="?", default="all", help="The target for yamk",
    )

    return parser.parse_args()


def main():
    args = parse_args()
    MakeCommand(args).make()



class FunctionMeta(type):
    functions = {}

    def __new__(mcs, name, bases, class_dict):
        x = super().__new__(mcs, name, bases, class_dict)
        if getattr(x, "name", None) is not None:
            mcs.functions[x.name] = x
        return x


class Function(metaclass=FunctionMeta):
    def __init__(self, base_dir):
        self.base_dir = base_dir


class Glob(Function):
    name = "glob"

    def __call__(self, *args):
        return [path.as_posix() for path in self.base_dir.glob(args[0])]


class Sort(Function):
    name = "sort"

    def __call__(self, *args):
        return sorted(*args)


class Exists(Function):
    name = "exists"

    def __call__(self, *args):
        return self.base_dir.joinpath(args[0]).exists()


class PWD(Function):
    name = "pwd"

    def __call__(self, *_args):
        return self.base_dir.as_posix()


class TernaryIf(Function):
    name = "ternary_if"

    def __call__(self, *args):
        return args[1] if args[0] else args[2]


class Substitute(Function):
    name = "sub"

    def __call__(self, *args):
        old, new, obj, *_ = args
        if isinstance(obj, list):
            return [self(old, new, string) for string in obj]
        return obj.replace(old, new)


functions = FunctionMeta.functions



import itertools
import os
import pathlib
import re
import subprocess
import sys

import tomlkit

from yamk import lib


class MakeCommand:
    def __init__(self, args):
        self.verbosity = args.verbose
        if self.verbosity:
            if self.verbosity > 1:
                print(args)
            sys.tracebacklimit = 9999
        self.regex_recipes = {}
        self.static_recipes = {}
        self.aliases = {}
        self.target = args.target
        makefile = args.makefile
        self.arg_vars = [
            {key: value}
            for key, value in map(
                lambda var: var.split("=", maxsplit=1), args.variables
            )
        ]
        self.base_dir = pathlib.Path(makefile).parent.absolute()
        self.phony_dir = self.base_dir.joinpath(".yamk")
        with open(makefile) as file:
            parsed_toml = tomlkit.parse(file.read())
        file_vars = parsed_toml.pop("$globals", {})
        self._parse_recipes(parsed_toml)
        self.vars = (
            lib.Variables(self.base_dir, **os.environ)
            .add_batch(self.arg_vars)
            .add_batch(file_vars.get("vars", []))
        )
        self.subprocess_kwargs = {"shell": True, "cwd": self.base_dir}

    def make(self):
        preprocessed = self._preprocess_target()
        for info in filter(
            lambda x: x["should_build"],
            sorted(preprocessed.values(), key=lambda x: x["priority"], reverse=True),
        ):
            self._make_target(info["recipe"])

    def _parse_recipes(self, parsed_toml):
        for target, raw_recipe in parsed_toml.items():
            if raw_recipe.get("alias"):
                self.aliases[target] = raw_recipe["alias"]
            recipe = lib.Recipe(target, raw_recipe, self.base_dir, self.arg_vars)

            if recipe.regex:
                self.regex_recipes[recipe.target] = recipe
            else:
                self.static_recipes[recipe.target] = recipe

    def _preprocess_target(self):
        recipe = self._extract_recipe(self.target)
        if recipe is None:
            raise ValueError(f"No recipe to build {self.target}")

        unprocessed = {self.target: {"recipe": recipe, "priority": 0}}
        preprocessed = {}
        while unprocessed:
            target, info = unprocessed.popitem()
            priority = info["priority"] + 1
            recipe = info["recipe"]
            preprocessed[target] = info
            for requirement in recipe.requires:
                recipe = self._extract_recipe(requirement)
                path = self._file_path(requirement)
                if requirement in preprocessed:
                    current_priority = preprocessed[requirement]["priority"]
                    preprocessed[requirement]["priority"] = max(
                        priority, current_priority
                    )
                elif recipe is not None:
                    unprocessed[requirement] = {
                        "recipe": recipe,
                        "priority": priority,
                    }
                elif path.exists():
                    preprocessed[requirement] = {"priority": priority}
                else:
                    raise ValueError(f"No recipe to build {requirement}")

        self._mark_unchanged(preprocessed)
        return preprocessed

    def _make_target(self, recipe):
        for command in recipe.commands:
            command, options = lib.extract_options(command)
            if recipe.echo or "echo" in options:
                print(command)
            result = subprocess.run(command, **self.subprocess_kwargs)
            if (
                result.returncode
                and not recipe.allow_failures
                and "allow_failures" not in options
            ):
                sys.exit(result.returncode)
        if recipe.phony and recipe.keep_ts:
            path = self._phony_path(recipe.target)
            self.phony_dir.mkdir(exist_ok=True)
            path.touch()

    def _extract_recipe(self, target):
        if target in self.aliases:
            target = self.aliases[target]

        absolute_path_target = self.base_dir.joinpath(target).as_posix()
        if target in self.static_recipes:
            recipe = self.static_recipes[target]
        elif absolute_path_target in self.static_recipes:
            recipe = self.static_recipes[absolute_path_target]
        else:
            for regex, recipe in self.regex_recipes.items():  # noqa: B007
                if re.fullmatch(regex, target) or re.fullmatch(
                    regex, absolute_path_target
                ):
                    break
            else:
                return None

        recipe.specify(target, self.vars)
        return recipe

    def _mark_unchanged(self, preprocessed):
        for target, info in sorted(
            preprocessed.items(), key=lambda x: x[1]["priority"], reverse=True
        ):
            ts = self._infer_timestamp(target, info)
            info["timestamp"] = ts
            should_build = self._should_build(target, preprocessed)
            info["should_build"] = should_build

    def _phony_path(self, target):
        encoded_target = target.replace(".", ".46").replace("/", ".47")
        return self.phony_dir.joinpath(encoded_target)

    def _file_path(self, target):
        return self.base_dir.joinpath(target)

    def _should_build(self, target, preprocessed):
        info = preprocessed[target]
        recipe = info.get("recipe")
        if recipe is None:
            return False
        if recipe.phony:
            if not recipe.keep_ts:
                return True
            path = self._phony_path(target)
            if not path.exists():
                return True
        else:
            path = self._file_path(target)
            if not path.exists():
                return True
            if recipe.exists_only:
                return False

        if any(
            preprocessed[requirement]["should_build"] for requirement in recipe.requires
        ):
            return True

        ts = info["timestamp"]
        req_ts = max(
            preprocessed[requirement]["timestamp"] for requirement in recipe.requires
        )
        return req_ts > ts

    def _infer_timestamp(self, target, info):
        recipe = info.get("recipe")
        path = self._file_path(target)
        if recipe is None:
            return path.stat().st_mtime
        if recipe.phony:
            path = self._phony_path(target)
            if recipe.keep_ts and path.exists():
                return path.stat().st_mtime
            return float("inf")
        if path.exists():
            if recipe.recursive:
                descendants = itertools.chain([path], path.rglob("*"))
                return max(p.stat().st_mtime for p in descendants)
            if recipe.exists_only:
                return 0
            return path.stat().st_mtime
        return float("inf")



import re
import shlex

from yamk.functions import functions

VAR = re.compile(
    r"(?P<dollars>\$+){(?P<variable>[a-zA-Z0-9_.]+)(?P<sep>:)?(?P<key>[a-zA-Z0-9_.]+)?}"
)
OPTIONS = re.compile(r"\[(?P<options>.*?)\](?P<string>.*)")
FUNCTION = re.compile(r"\$\(\((?P<name>\w+) *(?P<args>.*)\)\)")


class Recipe:
    def __init__(self, target, raw_recipe, base_dir, arg_vars):
        self._specified = False
        self.base_dir = base_dir
        self.vars: Variables
        self.arg_vars = arg_vars
        self.phony = raw_recipe.get("phony", False)
        self.requires = raw_recipe.get("requires", [])
        self.variable_list = raw_recipe.get("vars", [])
        self.commands = raw_recipe.get("commands", [])
        self.echo = raw_recipe.get("echo", False)
        self.regex = raw_recipe.get("regex", False)
        self.allow_failures = raw_recipe.get("allow_failures", False)
        self.target = self._target(target)
        if self.phony:
            self.keep_ts = raw_recipe.get("keep_ts", False)
            self.exists_only = False
            self.recursive = False
        else:
            self.keep_ts = False
            self.exists_only = raw_recipe.get("exists_only", False)
            self.recursive = raw_recipe.get("recursive", False)

    def __str__(self):
        if self._specified:
            return f"Specified recipe for {self.target}"
        return f"Generic recipe for {self.target}"

    def specify(self, target, globs):
        if self._specified:
            return
        self._specified = True
        if self.regex:
            groups = re.fullmatch(self.target, target).groupdict()
            self.target = target
        else:
            groups = {}
        self._update_variables(globs, groups)
        self._update_requirements()
        self._update_commands()

    def _update_variables(self, globs, groups):
        extra_vars = [groups, *self.arg_vars, {".target": self.target}]
        self.vars = globs.add_batch(self.variable_list).add_batch(extra_vars)

    def _update_requirements(self):
        parser = Parser(self.vars, self.base_dir)
        self.requires = parser.evaluate(self.requires)

    def _update_commands(self):
        extra_vars = [{".requirements": self.requires}]
        self.vars = self.vars.add_batch(extra_vars)
        parser = Parser(self.vars, self.base_dir)
        self.commands = parser.evaluate(self.commands)

    def _target(self, target):
        if not self.phony:
            target = self.base_dir.joinpath(target).as_posix()
        if self.regex:
            target = re.compile(target)
        return target


class Variables(dict):
    def __init__(self, base_dir, **kwargs):
        super().__init__(**kwargs)
        self.base_dir = base_dir

    def add_batch(self, batch):
        new_vars = self.__class__(self.base_dir, **self)
        for var_block in batch:
            parser = Parser(new_vars, self.base_dir)
            for key, value in var_block.items():
                key = parser.evaluate(key)
                key, options = extract_options(key)
                if "weak" in options:
                    continue
                value = parser.evaluate(value)
                new_vars[key] = value
        return new_vars


class Parser:
    def __init__(self, variables, base_dir):
        self.vars = variables
        self.base_dir = base_dir

    @staticmethod
    def _stringify(value):
        if isinstance(value, list):
            return " ".join(map(str, value))
        return str(value)

    def expand_function(self, name, args):
        args = shlex.split(args)
        for i, arg in enumerate(args):
            args[i] = self.evaluate(arg)
        function = functions.get(name)(self.base_dir)
        return function(*args)

    def repl(self, matchobj):
        dollars = matchobj.group("dollars")
        variable = matchobj.group("variable")
        key = matchobj.group("key")
        if len(dollars) % 2:
            value = self.vars[variable]
            if key is None:
                return self._stringify(value)
            if isinstance(value, list):
                key = int(key)
            return self._stringify(value[key])
        return f"{'$'*(len(dollars)//2)}{{{variable}}}"

    def substitute(self, string):
        function = re.fullmatch(FUNCTION, string)
        if function is not None:
            return self.expand_function(**function.groupdict())

        if (
            string.startswith("$")
            and not string.startswith("$$")
            and re.fullmatch(VAR, string)
        ):
            match = re.fullmatch(VAR, string)
            if match.group("sep") is None:
                return self.vars[match.group("variable")]
        return re.sub(VAR, self.repl, string)

    def evaluate(self, obj):
        if isinstance(obj, str):
            return self.substitute(obj)
        if isinstance(obj, list):
            out = []
            for string in obj:
                evaluated = self.evaluate(string)
                if isinstance(evaluated, list):
                    out.extend(evaluated)
                else:
                    out.append(evaluated)
            return out
        if isinstance(obj, dict):
            return {
                self.evaluate(key): self.evaluate(value) for key, value in obj.items()
            }
        raise TypeError(f"{obj.__class__.__name__} is not supported for evaluation")


def extract_options(string):
    match = re.fullmatch(OPTIONS, string)
    if match is None:
        return string.strip(), set()

    options = match.group("options")
    string = match.group("string")
    return (
        string,
        set(map(lambda s: s.strip(), options.split(","))),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['yamk']

package_data = \
{'': ['*']}

install_requires = \
['tomlkit>=0.5.11,<0.6.0']

entry_points = \
{'console_scripts': ['yam = yamk.main:main', 'yamk = yamk.main:main']}

setup_kwargs = {
    'name': 'yamk',
    'version': '0.9.1',
    'description': 'Yet another make',
    'long_description': '<p align="center">\n<a href="https://travis-ci.org/spapanik/yamk"><img alt="Build" src="https://travis-ci.org/spapanik/yamk.svg?branch=master"></a>\n<a href="https://coveralls.io/github/spapanik/yamk"><img alt="Coverage" src="https://coveralls.io/repos/github/spapanik/yamk/badge.svg?branch=master"></a>\n<a href="https://github.com/spapanik/yamk/blob/master/LICENSE.txt"><img alt="License" src="https://img.shields.io/github/license/spapanik/yamk"></a>\n<a href="https://pypi.org/project/yamk"><img alt="PyPI" src="https://img.shields.io/pypi/v/yamk"></a>\n<a href="https://pepy.tech/project/yamk"><img alt="Downloads" src="https://pepy.tech/badge/yamk"></a>\n<a href="https://github.com/psf/black"><img alt="Code style" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>\n\n\n# yamk: yet another make\n\n`yamk` offers an alternative tool to control the housekeeping tasks of a project, as well as the creation of executables and non-source files from source files.\n\n## Installation\n\nThe easiest way is to use pip to install `yamk`.\n\n```bash\npip install --user yamk\n```\n\nPlease make sure that the correct directory is added to your path. This depends on the OS.\n\n## Usage\n`yamk`\'s behaviour is defined in a toml file. The expected name is `make.toml`, but you can specify a different file if you want.\n\n`yamk` can be invoked by using the command `yamk`, which is also aliased to `yam`. `yamk` follows the GNU recommendations for command line interfaces, and offers:\n*  -h/--help              show this help message and exit\n*  -V/--version           print the version and exit\n*  -m/--makefile MAKEFILE the path to makefile (defaults to `./make.toml`)\n\n## Terminology\n\n* **Target**: A target is the file that needs to be created, or the task that should be done. The first type of targets is called a file target, and the second type, a phony target. A file target, is a file in the UNIX sense of what a file is.\n* **Recipe**: A recipe is what tells `yamk` how to build a target, or do the task specified by the target.\n* **Strength**: Whenever there is ambiguity in the variables or targets the strength of the variable or the target will be what will decide which will be executed.\n\n## Recipes\n\nA recipe has the following format:\n```toml\n[target_name]\n<key_1> = <value_1>\n<key_2> = <value_2>\n...\n<key_n> = <value_n>\n```\n\n## Targets\n\nThere are two groups of mutually exclusive types of targets: File/phony targets and static/regex targets. There are also two more cases, but they are not targets: Aliases and meta-targets.\n\n### Meta-targets\nA target starting with a single dollar sign is reserved by `yamk` itself, for meta-targets. If the target starts with more than one dollar sign, a single dollar sign is stripped, and it\'s a normal target or an alias. Currently the only meta-target is `$globals`, that has one valid key: `vars`. `vars` is the same as the key with the same name if static and regex targets, but weaker.\n\n### Aliases\nAn alias has a single key, named `alias`, and the value of the key is the name of the target it aliases. \n\n### File targets\nA file target is a target who doesn\'t have the `phony` key, or that the `phony` key is specifically set to a false value. The path can be with its absolute path, or the path relative to the makefile. Absolute paths are discouraged, as they limit the users ability to choose.\n\n### Phony targets\nA target that contains the `phony` key set to a true value, is a phony target.\n\n### Static targets\nA static target is a target who doesn\'t have the `regex` key, or that the `regex` key is specifically set to a false value.\n\n### Regex targets\nA target that contains the `regex` key set to a true value, is a regex target. A regex target is assumed to be a file target. In case the target matches a regex target, any name that was matched via a named grouped will be available as a variable. Regex targets are weaker than static ones.\n\n## Keys\n\nThe following list contains all the valid keys. The list is in the form <key_name>: <value_type> (target types). A target type starting with a tilde(~) is an invalid target, and "any" means that this key is valid for all target types. Adding an invalid key is undefined behaviour that will most probably be ignored, or crash.\n\n##### phony: boolean (~regex targets)\nIf set to true, it turns the target to a phony target.\n\n##### regex: boolean (~phony targets)\nIf set to true, it turns the target to a regex target.\n\n##### requires: array of strings (any)\nAn array of all the requirements for the target. The requirements that need to be built, are guaranteed to be built before the target, but there are no guarantees on the order that the requirements will be built.\n\n##### commands: array of strings (any)\nThe commands to build the target.\n\n##### vars: array of tables (any)\nThe variables specific to this target.\n\n##### keep_ts: boolean (phony targets)\nIf set to true, this phony target will keep the timestamp when it was last built. In general, a phony target will be built every time, but a phony target with `keep_ts` set to true, will only be built if the kept timestamp is newer than the timestamp of the requirements.\n\n##### exists_only: boolean (file targets)\nIf set to true, this file will be built only if it doesn\'t exist. The check for the existence is happening before any make command is run and there is a de-duplication step, but some care should be taken into account if it\'s created as a side-effect of another target.\n\n##### recursive: boolean (file targets)\nThis makes sense only for directories. In this case, if it\'s set to true, the timestamp of the directory is taken to be the maximum of all the timestamps of all files, at any depth, inside it.\n\n##### echo: boolean (any)\nBy default, the commands are not echoed before they are run. If set to true, they do.\n\n##### allow_failures: boolean (any)\nBy default, a failing command will halt the execution of the makefile. This will allow `yamk` to try and recover from the error. Allow_failures doesn\'t guarantee that the execution will resume, because the nature of the failure may not allow that.\n\n## Variables\n\nA variable is defined in the key `vars`. The variables are parsed in order, so a variable, once defined can be used later. An example of `vars`:\n\n```toml\n[target_name]\nvars = [\n    {"[variable_options]var_1": <value_1>},\n    {"[variable_options]var_2": <value_2>},\n    ...\n    {"[variable_options]var_n": <value_n>},\n]\n```\n\nThe options are completely optional, and they are comma separated.\n\nA variable that resolves to a list or a dictionary will be treated as such. If it should be treated as a string instead instead of "${variable}", the form "${variable:}" should be used. If a variable is a part of a string, it will always be transformed into a string first. A key can be passed to pick up a specific value from a list or a dictionary, for example, "${list_var:0}" or "${dict_var:key}".\n\n### Variable types\n\nThere are six types of variables: environment variables, argument, global, local, regex and implicit ones. With an increasing strength order:\n\n##### Environment\nAll the environment variables gathered at the beginning of the execution of `yamk` are gathered into variables. They are the weakest variables.\n\n##### Global\nA global variable is a variable specified in the `$global` meta-target.\n\n##### Local\nA local variable is a variable specified by the `vars` key inside a target.\n\n##### Regex\nIn the case of a regex target, any named group is a regex variable, which has the value of the matched text.\n\n##### Argument\nAn argument variable is a variable defined with the --variable option when `yamk` was invoked.\n\n##### Implicit\nThe implicit variables are variables created by the target itself, implicitly. They are guarded against overriding, as they start with a dot. This dot acts as a safeguard so they cannot mix with environment and regex ones. At the moment, the following two implicit variables exist:\n* `.target`: the name of the target. In case of a file target, it\'s the absolute path to the file, regardless of they way it was defined.\n* `.requirements`: the array of the requirements. All the file requirements are given as their absolute paths.\nAll the implicit variables, they are specific to the target that it\'s being built, i.e. if the file target is specified as a regular expression, the absolute path to the specific file, not the regex path.\n\n### Variable options\n\nThe only option at the moment is `weak`.\n\n##### Weak\nWeak can be used to make the variable keep its value if it\'s not unset\n\n## Commands\n\n### Command options\n\nThere are two command options: `echo` and `allow_failures`. These commands can be used to customise the specific command, as if the respective variable was set.\n',
    'author': 'Stephanos Kuma',
    'author_email': 'spapanik21@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/spapanik/yamk',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.0,<4.0.0',
}


setup(**setup_kwargs)
